#ifndef _AI_H
#define _AI_H

#include "carinput.h"

#include <vector>

class CAR;
class BEZIER;
class TRACK;

struct AI_Car
{
	AI_Car (CAR * new_car) : car (new_car), shift_time(0.0),
		longitude_mu(0.9), lateral_mu(0.9),last_patch(NULL)
		{
			inputs.resize(CARINPUT::INVALID, 0.0);
		}
	CAR * car;
	float shift_time;
	float longitude_mu; //friction coefficient of the tire - longitude direction
	float lateral_mu; //friction coefficient of the tire - lateral direction
	const BEZIER * last_patch; //last patch the car was on, used in case car is off track
	
	///the vector is indexed by CARINPUT values
	std::vector <float> inputs;
};
 
class AI
{
private:
	std::vector <float> empty_vector;
	std::vector <AI_Car> AI_Cars;
	void updateGasBrake(AI_Car *c, float dt, TRACK* track_p);
	void calcMu(AI_Car *c, TRACK* track_p);
	float calcSpeedLimit(AI_Car *c, const BEZIER* patch, float friction);
	float calcBrakeDist(AI_Car *c, float current_speed, float allowed_speed, float friction);
	//void updateGear(AI_Car *c, float dt);
	//float DownshiftPoint(AI_Car *c, int gear);
	//void updateClutch(AI_Car *c, float dt);
	void updateSteer(AI_Car *c, float dt);
	double Angle(double x1, double y1); ///< returns the angle in degrees of the normalized 2-vector


public:
	void add_car(CAR * car);
	void clear_cars() { AI_Cars.clear(); }
	void update(float dt, TRACK* track_p);
	const std::vector <float> & GetInputs(CAR * car) const; ///< returns an empty vector if the car isn't AI-controlled
};

#endif //_AI_H
