#include "collision_detection.h"

#include <list>
using std::list;

void COLLISION_WORLD::Init()
{
	if (id != NULL || collisionconfig != NULL || collisiondispatcher != NULL || collisionbroadphase != NULL)
		DeInit();
	
	collisionconfig = new btDefaultCollisionConfiguration();
	collisiondispatcher = new btCollisionDispatcher(collisionconfig);
	
	btVector3 worldAabbMin(-10000,-10000,-10000);
	btVector3 worldAabbMax(10000,10000,10000);
	
	//collisionbroadphase = new bt32BitAxisSweep3(worldAabbMin,worldAabbMax);
	collisionbroadphase = new btSimpleBroadphase;
	
	id = new btCollisionWorld(collisiondispatcher,collisionbroadphase,collisionconfig);
}

void COLLISION_WORLD::DeInit()
{
	if (id != NULL)
		delete id;
	id = NULL;
	
	if (collisionbroadphase != NULL)
		delete collisionbroadphase;
	collisionbroadphase = NULL;
	
	if (collisiondispatcher != NULL)
		delete collisiondispatcher;
	collisiondispatcher = NULL;
	
	if (collisionconfig != NULL)
		delete collisionconfig;
	collisionconfig = NULL;
}

void COLLISION_WORLD::CollideRay(const MATHVECTOR <float, 3> & position, const MATHVECTOR <float, 3> & direction, const float length, std::list <COLLISION_CONTACT> & outputcontactlist, const COLLISION_SETTINGS & settings) const
{
	outputcontactlist.clear();
	
	//return;
	
	assert (id); //Tried to collide objects before physics was initialized
	
	MATHVECTOR <float, 3> rayend = position + direction * length;
	
	btVector3 rayFrom(position[0], position[1], position[2]);
	btVector3 rayTo (rayend[0], rayend[1], rayend[2]);
	
	MultipleRayResultCallback colresult(rayFrom, rayTo);
	
	short int raymask = 0;
	if (settings.GetMapCollide())
		raymask |= 2;
	if (settings.GetPlayerCollide())
		raymask |= 4;
	if (settings.GetRoadCollide())
		raymask |= 8;
	
	//raymask = 255;
	
	list <COLLISION_OBJECT *> candidates;
	AABB <float> raybox;
	MATHVECTOR <float, 3> rpos = position;
	MATHVECTOR <float, 3> rdir = direction;
	/*if (settings.GetMapCollide())
	{
		MATHVECTOR <float, 3> rstart(game.objects.GetSceneNode().TransformIntoLocalSpace(position));
		MATHVECTOR <float, 3> rend(game.objects.GetSceneNode().TransformIntoLocalSpace(rayend));
		raybox.SetFromCorners(rstart, rend);
		
		rpos = game.objects.GetSceneNode().TransformIntoLocalSpace(position);
		rdir = game.objects.GetSceneNode().TransformIntoLocalSpace(direction);
	}
	else*/
	{
		raybox.SetFromCorners(position,rayend);
		//cout << "test" << endl;
	}
	//colspeedup.Query(position, direction, length, candidates);
	colspeedup.Query(rpos, rdir, length, candidates);
	//colspeedup.Query(raybox, candidates);
	
	btTransform rayFromTrans,rayToTrans;
	rayFromTrans.setIdentity();
	rayFromTrans.setOrigin(rayFrom);
	rayToTrans.setIdentity();
	rayToTrans.setOrigin(rayTo);
	
	//cout << "collision candidates: " << candidates.size() << endl;
	int count(1);
	for (list <COLLISION_OBJECT *>::iterator i = candidates.begin(); i != candidates.end(); i++)
	{
		btCollisionObject * collisionObject = &((*i)->GetBulletObject());
		
		//if (collisionObject->getBroadphaseHandle()->m_collisionFilterGroup & raymask)
		if ((settings.GetMapCollide() && (*i)->IsMapObject()) ||
			(settings.GetRoadCollide() && (*i)->IsRoadObject()) ||
			(settings.GetPlayerCollide() && (*i)->IsPlayerObject()))
		{
			id->rayTestSingle(rayFromTrans,rayToTrans,
					collisionObject,
       					collisionObject->getCollisionShape(),
					collisionObject->getWorldTransform(),
					colresult);
		}
		
		count++;
	}
	
	//id->rayTest(rayFrom, rayTo, colresult, raymask);
	
	//cout << "Ray collisions: " << colresult.results.size() << endl;
	
	if (!colresult.results.empty())
	{
		//HandleCollision(settings, &colresult, outputcontactlist);
		
		for (list <btCollisionWorld::LocalRayResult>::iterator i = colresult.results.begin(); i != colresult.results.end(); i++)
		{
			//cout << "processing collision" << endl;
			
			assert (i->m_collisionObject); //Collision hit isn't tied to an object
			
			if (PassesFilter(settings, i->m_collisionObject->getUserPointer()))
			{
				COLLISION_CONTACT newcont;
				outputcontactlist.push_back(newcont);
				
				MATHVECTOR <float, 3> normal;
				normal.Set(i->m_hitNormalLocal.x(),i->m_hitNormalLocal.y(),i->m_hitNormalLocal.z());
				normal = normal.Normalize();
				
				btVector3 hp;
				hp.setInterpolate3(colresult.m_rayFromWorld,colresult.m_rayToWorld,i->m_hitFraction);
				MATHVECTOR <float, 3> hpv;
				hpv.Set(hp.x(),hp.y(),hp.z());
				outputcontactlist.back().Set(hpv,
					normal,
					(colresult.m_rayToWorld - colresult.m_rayFromWorld).length() * i->m_hitFraction,
					(COLLISION_OBJECT*)i->m_collisionObject->getUserPointer(),
					(COLLISION_OBJECT*)i->m_collisionObject->getUserPointer());
				
				//outputcontactlist.pop_back();
				
				//cout << "PHYSICS::COLLIDERAY Collision hit: raylength=" << (colresult.m_rayToWorld - colresult.m_rayFromWorld).length() << ", hitfraction=" << i->m_hitFraction << ", depth=" << (colresult.m_rayToWorld - colresult.m_rayFromWorld).length() * i->m_hitFraction << ", hitpoint=";VERTEX(hp.x(),hp.y(),hp.z()).DebugPrint();
				//cout << "Collision hit: raylength=" << (colresult.m_rayToWorld - colresult.m_rayFromWorld).length() << ", hitfraction=" << i->m_hitFraction << ", depth=" << (colresult.m_rayToWorld - colresult.m_rayFromWorld).length() * i->m_hitFraction << ", normal=";normal.DebugPrint();
			}
			//else cout << "discarded collision due to filtering" << endl;
		}
	}
}

void COLLISION_WORLD::CollideBox(const MATHVECTOR <float, 3> & position, const QUATERNION <float> & orientation, const MATHVECTOR <float, 3> & dimensions, std::list <COLLISION_CONTACT> & outputcontactlist, const COLLISION_SETTINGS & settings) const
{
	outputcontactlist.clear();
	
	//return;
	
	assert(id); //Tried to collide objects before physics was initialized
	
	btCollisionObject object;
	object.getWorldTransform().setIdentity();
	object.getWorldTransform().setOrigin(btVector3(position[0],position[1],position[2]));
	object.getWorldTransform().setRotation(btQuaternion(orientation.x(),orientation.y(),orientation.z(),orientation.w()));
	btBoxShape box(btVector3(dimensions[0]*0.5,dimensions[1]*0.5,dimensions[2]*0.5));
	object.setCollisionShape(&box);
	
	short int mask = 0;
	if (settings.GetMapCollide())
		mask |= 2;
	if (settings.GetPlayerCollide())
		mask |= 4;
	if (settings.GetRoadCollide())
		mask |= 8;
	
	//id->addCollisionObject(&object, 1, mask);
	
	//determine collision candidates in a speedy way
	list <COLLISION_OBJECT *> candidates;
	AABB <float> querybox;
	//VERTEX boxpos1(position-dimensions.ScaleR(0.5));
	//VERTEX boxpos2(position+dimensions.ScaleR(0.5));
	MATHVECTOR <float, 3> boxpos1, boxpos2;
	btVector3 aabbmin, aabbmax;
	box.getAabb(object.getWorldTransform(), aabbmin, aabbmax);
	boxpos1.Set(aabbmin.x(),aabbmin.y(),aabbmin.z());
	boxpos2.Set(aabbmax.x(),aabbmax.y(),aabbmax.z());
	querybox.SetFromCorners(boxpos1,boxpos2);

	colspeedup.Query(querybox, candidates);
	
	//if (!candidates.empty()) cout << candidates.size() << endl;
	
	for (list <COLLISION_OBJECT *>::iterator i = candidates.begin(); i != candidates.end(); i++)
	{
		btCollisionAlgorithm * algo;
		algo = id->getDispatcher()->findAlgorithm(&object, &((*i)->GetBulletObject()));
		
		assert (algo); //Couldn't find collision algorithm
		
		btManifoldResult manifoldout(&object, &((*i)->GetBulletObject()));
		algo->processCollision(&object, &((*i)->GetBulletObject()), id->getDispatchInfo(), &manifoldout);
	}
	
	//find the number of hits
	int num_manifolds = id->getDispatcher()->getNumManifolds();
	
	//if (num_manifolds > 0) cout << num_manifolds << " manifolds" << endl;
	
	for (int i=0; i < num_manifolds; i++)
	{
		//find the number of contacts
		btPersistentManifold* manifold = id->getDispatcher()->getManifoldByIndexInternal(i);
		btCollisionObject* obA = static_cast<btCollisionObject*>(manifold->getBody0());
		btCollisionObject* obB = static_cast<btCollisionObject*>(manifold->getBody1());
		manifold->refreshContactPoints(obA->getWorldTransform(),obB->getWorldTransform());
		
		btCollisionObject * otherobject(NULL);
		bool obAisbox(obA == &object);
		
		if (obAisbox)
			otherobject = obB;
		else
			otherobject = obA;
		
		int num_contacts = manifold->getNumContacts();
		//if (num_contacts > 0) cout << i << ". " << num_contacts << " contacts" << endl;
		for (int j=0; j < num_contacts; j++)
		{
			btManifoldPoint& pt = manifold->getContactPoint(j);
			btVector3 ptA = pt.getPositionWorldOnA();
			btVector3 ptB = pt.getPositionWorldOnB();
			MATHVECTOR <float, 3> boxpt, otherpt;
			
			if (obAisbox)
			{
				boxpt.Set(ptA.x(), ptA.y(), ptA.z());
				otherpt.Set(ptB.x(), ptB.y(), ptB.z());
			}
			else
			{
				otherpt.Set(ptA.x(), ptA.y(), ptA.z());
				boxpt.Set(ptB.x(), ptB.y(), ptB.z());
			}

			if (PassesFilter(settings, otherobject->getUserPointer()))
			{
				COLLISION_CONTACT newcont;
				outputcontactlist.push_back(newcont);
				
				MATHVECTOR <float, 3> ptv;
				ptv.Set(pt.m_normalWorldOnB.x(),pt.m_normalWorldOnB.y(),pt.m_normalWorldOnB.z());
				
				outputcontactlist.back().Set(otherpt,
						ptv,
						(otherpt-boxpt).Magnitude(),
						(COLLISION_OBJECT*)otherobject->getUserPointer(),
						(COLLISION_OBJECT*)otherobject->getUserPointer());
			}
		}
		
		//manifold->clearManifold();
	}
	
	for (int i=0; i < num_manifolds; i++)
		id->getDispatcher()->releaseManifold(id->getDispatcher()->getManifoldByIndexInternal(0));
}

bool COLLISION_WORLD::PassesFilter(const COLLISION_SETTINGS & settings, void * checkme) const
{
	bool exception = false;
	
	//int count = 1;
	for (list <void *>::const_iterator i = settings.GetExceptionObjectIDs().begin(); i != settings.GetExceptionObjectIDs().end(); i++)
	{
		//cout << count << ". " << *i << " =? " << checkme;
		exception = exception || (*i == checkme);
		
		//if (exception)
		//	cout << " (exception)";
		//cout << endl;
		//++count;
	}
	
	return !exception;
}

void COLLISION_OBJECT::InitTrimesh(const float * vertices, int vstride, int vcount, const int * faces, int fcount, int istride, const float * normals, const COLLISION_OBJECT_SETTINGS & objsettings)
{
	assert(!loaded); //Tried to double load a physics object
	
	if (trimesh != NULL || trimesh_varray != NULL)
		DeInit();
	
	assert(fcount % 3 == 0); //Face count is not a multiple of 3
	
	trimesh_varray = new btTriangleIndexVertexArray(fcount/3, (int*) faces, istride, vcount, (float*) vertices, vstride);
	
	trimesh = new btBvhTriangleMeshShape(trimesh_varray, true);
	
	//std::cerr << trimesh_varray << "," << trimesh << std::endl;
	
	id.getWorldTransform().setIdentity();
	
	id.setCollisionShape(trimesh);
	
	btVector3 AabbMin,AabbMax;
	id.getCollisionShape()->getAabb(id.getWorldTransform(),AabbMin,AabbMax);
	MATHVECTOR <float, 3> bboxmin;
	bboxmin.Set(AabbMin.x(),AabbMin.y(),AabbMin.z());
	MATHVECTOR <float, 3> bboxmax;
	bboxmax.Set(AabbMax.x(),AabbMax.y(),AabbMax.z());
	
	bbox.SetFromCorners(bboxmin, bboxmax);
	
	//colspeedup->Add(this);
	
	//group = 255;
	//mask = 255;
	
	id.setUserPointer(const_cast<void *>(objsettings.ObjID()));
	//id.setUserPointer(this);
	//cout << "set physics object user pointer to: " << id.getUserPointer() << endl;
	//cout << "set physics object user pointer to: " << settings.ObjID() << endl;
	
	//worldid->addCollisionObject(&id, group, mask);
	
	settings = objsettings;
	
	loaded = true;
}

void COLLISION_OBJECT::InitTrimesh(const VERTEXARRAY & varray, const COLLISION_OBJECT_SETTINGS & objsettings)
{
	const float * vertices;
	int vcount;
	const int * faces;
	int fcount;
	float * normals(NULL);
	varray.GetVertices(vertices, vcount);
	varray.GetFaces(faces, fcount);
	//std::cout << "verts: " << vcount << ", faces: " << fcount << std::endl;
	InitTrimesh(vertices, sizeof(float)*3, vcount, faces, fcount, sizeof(int)*3, normals, objsettings);
}

void COLLISION_OBJECT::DeInit()
{
	if (trimesh_varray != NULL)
		delete trimesh_varray;
	trimesh_varray = NULL;
	
	if (trimesh != NULL)
		delete trimesh;
	trimesh = NULL;
}

void COLLISION_OBJECT::SetPosition(const MATHVECTOR <float, 3> & newpos)
{
	assert(loaded); //Physics object not loaded yet
	id.getWorldTransform().setOrigin(btVector3(newpos[0],newpos[1],newpos[2]));
}

void COLLISION_OBJECT::SetQuaternion(const QUATERNION <float> & newquat)
{
	assert(loaded); //Physics object not loaded yet
	id.getWorldTransform().setRotation(btQuaternion(newquat.x(),newquat.y(),newquat.z(),newquat.w()));
}

bool COLLISION_CONTACT::CollideRay(const MATHVECTOR <float, 3> & origin, const MATHVECTOR <float, 3> & direction, const float length, COLLISION_CONTACT & output_contact) const
{
	/*//simple approximation
	output_contact.Set(position, normal, depth, col1, col2);
	return true;*/

	//enhanced plane-based approximation
	float D = - normal.dot ( position );
	float Pn_dot_Rd = normal.dot ( direction );
	bool newcon ( true );
	if ( Pn_dot_Rd != 0 )
	{
		float t = - ( normal.dot ( origin ) + D ) / ( Pn_dot_Rd );
		if ( t >= 0 )
		{
			MATHVECTOR <float, 3> newpos = origin + direction*t;

			float newdepth = t;

			output_contact.Set(newpos, normal, newdepth, col1, col2);
		}
		else
			newcon = false;
	}
	else
		newcon = false;
	
	return newcon;
}
